/** @defgroup automaton automaton classes
 *  @brief automaton (deterministic or non-deterministic, probabilistic or not)
 */
// @{

#ifndef __AUTOMATON_H__
#define __AUTOMATON_H__

//STL
#include <vector>
#include <queue>
#include <stack>
#include <functional>
#include <map>
//STD
#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <string>
//STR
using namespace std;
#include "macro.h"
#include "seed.h"

#ifndef NOMATRIX
#include "matrix.hh" 
#endif

class automaton;

/** Define the product families of operators
 */

typedef enum ProductType {
  PRODUCT_UNION_FINAL_LOOP,
  PRODUCT_UNION_NO_FINAL_LOOP,
  PRODUCT_UNION_NO_FINAL_LOOP_ADD,
  PRODUCT_INTERSECTION_FINAL_LOOP,
  PRODUCT_INTERSECTION_NO_FINAL_LOOP,
  PRODUCT_BUTNOT_FINAL_LOOP,
  PRODUCT_BUTNOT_NO_FINAL_LOOP,
  PRODUCT_NOTBUT_FINAL_LOOP,
  PRODUCT_NOTBUT_NO_FINAL_LOOP,
} ProductType;


/// final
#define TRUE  1
/// non final
#define FALSE 0


/**
 * @class transition
 *      
 * @brief transition to a given state (on a given letter)
 */

class transition {
 public:
  /** @brief build a transition object
   *  @param state gives the state number reached by this transition
   *  @param prob  gives the probability of this transition
   */
 transition(int state = 0, double prob = 1e+0) : _state(state), _prob(prob) {};

 protected :
  /// state number to be reached
  int     _state;
  /// transition probability (1.0 if deterministic automaton) 
  double  _prob;

  /// state is a friend class to ease access
  friend class state;
  /// automaton is a friend class to ease access
  friend class automaton;
  /// print automaton information is friend
  friend ostream& operator<<(ostream& os, const automaton& automaton);
  /// read  automaton information
  friend istream& operator>>(istream& is, automaton& automaton);
};

/**
 * @class state
 *      
 * @brief state represented as a table of list of transitions ("next[gv_align_alphabet_size]") + a table of lists of back transitions ("prev[gv_align_alphabet_size]") 
 *
 */

class state {
 public:
  /// Build an empty state (empty transition lists)
 state(int final = 0): _final(final) { _next  =  vector< vector < transition > >(gv_align_alphabet_size, vector < transition > ());};
  

  /// Erase a state (clear transition lists first)
  ~state() { for(int a = 0; a < gv_align_alphabet_size; a++) _next[a].clear();  _next.clear(); };
  
  /** @brief Clear transition lists
   */
  void clear() { for(int a = 0 ; a < gv_align_alphabet_size ; a++) _next[a].clear();  _next.clear(); };

 protected:
  /// forward  transitions list on letter A
  vector < vector < transition > >  _next;
  /// final state
  int _final;

  /// automaton is a friend class to ease access
  friend class automaton;
  /// print automaton information is friend
  friend ostream& operator<<(ostream& os, const automaton& automaton);
  /// read  automaton information
  friend istream& operator>>(istream& is, automaton& automaton);

};




/**
 * @class automaton
 *
 * @brief automaton (deterministic or non-deterministic, probabilistic or not) roughly represented as a vector of states
 *
 *  this one can either be affected to: 
 *      @li a seed automaton  (deterministic, initial state is 1, final states have the final tag)
 *      @li a probability model (Bernoulli, Markov, and old M3,M14,HMM models, can either be deterministic or non-deterministic)
 *      @li a target alignment model (deterministic)
 *
 */

class automaton {
 public:
  
  /// constructor for an empty automaton
  automaton() {
         _states = vector<state>(0);
    _init_states = vector<int>(0);
  };

  /// destructor for all possibles automata build by this class
  ~automaton() {
    for(int i = 0 ; i < (int)_states.size() ; i++){
      _states[i].clear();
    }
         _states.clear();
    _init_states.clear();
  };
  
  /** @name  Build automata 
   *  @brief each one builds a special automaton 
   */

  // @{
  
  /** @brief Old algorithm in @f$ \mathcal{O}(w 2^{s-w})@f$ for subset seed matching (use on index of size @f$ \mathcal{O}(w 2^{s-w})@f$ )
   *  @param s is the seed descriptor
   *  @param matchingmatrix is a boolean  matrix that gives for alignment letter "a", and seed letter "b", the matching with "matrix[a][b]".
   *  @param nomerge if final states dont have to be merged together into a single state (default behaviour is merging).
   *  @return 0 
   */

  int Automaton_SeedPrefixesMatching_old(const seed & s,
                                         const vector< vector <int> > & matchingmatrix,
                                         const bool nomerge=false);


  /** @brief New linear algorithm for subset seed matching (CIAA 2007)
   *  @param s is the seed descriptor
   *  @param matchingmatrix is a boolean  matrix that gives for alignment letter "a", and seed letter "b", the matching with "matrix[a][b]".
   *  @param nomerge if final states dont have to be merged together into a single state (default behaviour is merging).
   *  @return 0 
   */

  int Automaton_SeedPrefixesMatching    (const seed & s,
                                         const vector< vector <int> > & matchingmatrix,
                                         const bool nomerge=false);


  /** @brief Modified version that takes lossless costs into account
   *  @param s is the seed descriptor
   *  @param matchingmatrix is a boolean  matrix that gives for alignment letter "a", and seed letter "b", the matching with "matrix[a][b]".
   *  @param nomerge if final states dont have to be merged together into a single state (default behaviour is merging).
   *  @param costs gives a vector of costs for each align alphabet letter
   *  @param cost_threshold gives a cost threshold that must not be reached by any alignment.
   *  @return 0 
   *  @see Automaton_SeedPrefixesMatching
   */

  int Automaton_SeedPrefixesMatchingCost (const seed& s,
                                          const vector< vector <int> > & matchingmatrix,
                                          const bool nomerge,
                                          const vector<int> & costs,
                                          const int cost_threshold);

  
  /** @brief Aho-Corasick automaton used by J.Buhler
   *  @param s is the seed descriptor
   *  @param matchingmatrix is a boolean  matrix that gives for alignment letter "a", and seed letter "b", the matching with "matrix[a][b]".
   *  @param nomerge if final states dont have to be merged together into a single state (default behaviour is merging).
   *  @return 0 
   */
  int Automaton_SeedBuhler(const seed & s, 
                           const vector< vector <int> > & matchingmatrix, 
                           const bool nomerge=false);
  
  /** @brief Aho-Corasick automaton with scoring method to prune prefixes 
   *  @param s is the seed descriptor
   *  @param matchingmatrix is a boolean  matrix that gives for alignment letter "a", and seed letter "b", the matching with "matrix[a][b]".
   *  @param scoringmatrix is an integer matrix that gives for alignment letter "a", and seed letter "b", the score    with "matrix[a][b]".
   *  @param scoringthreehold is the minimal score that has to be reached to have a match for the seed.
   *  @param nomerge if final states dont have to be merged together into a single state (default behaviour is merging).
   *  @return 0 
   */

  int Automaton_SeedScore    (const seed & s,  
                              const vector< vector <int> > & matchingmatrix, 
                              const vector< vector<int> >  & scoringmatrix, 
                              const int scoringthreehold,
                              const bool nomerge=false);


  /** @brief Modified version that takes lossless costs into account
   *  @param s is the seed descriptor
   *  @param matchingmatrix is a boolean  matrix that gives for alignment letter "a", and seed letter "b", the matching with "matrix[a][b]".
   *  @param scoringmatrix is an integer matrix that gives for alignment letter "a", and seed letter "b", the score    with "matrix[a][b]".
   *  @param scoringthreehold is the minimal score that has to be reached to have a match for the seed.
   *  @param nomerge if final states dont have to be merged together into a single state (default behaviour is merging).
   *  @param costs gives a vector of costs for each align alphabet letter
   *  @param cost_threshold gives a cost threshold that must not be reached by any alignment.
   *  @return 0 
   *  @see Automaton_SeedScore
   */

  int Automaton_SeedScoreCost(const seed & s, 
                              const vector< vector <int> > & matchingmatrix, 
                              const vector< vector<int> > & scoringmatrix, 
                              const int scoringthreehold, 
                              const bool nomerge, 
                              const vector<int> & costs, 
                              const int cost_threshold);

   
  /** @brief build a probabilistic bernoulli model
   *  @param p gives the probability of each letter of A (sum must be equal to 1.0).
   *  @return 0 
   */

  int Automaton_Bernoulli(const vector<double> & p);


  /** @brief build a probabilistic markov model of order k
   *  @param p gives the probability of each word of A^k (sum must be equal to 1.0).
   *  @param k gives the order.
   *  @return 0 
   */

  int Automaton_Markov(const vector<double> & p, 
                       const int k);

  /** @brief build an Homogeneous sequence automaton. 
   *  @details It represents an alignment such that no substring of the alignment has a score greater than the full alignment
   *  @param scores gives a table of scores for each of the A letters.
   *  @param length gives the alignment length.
   *  @return 0 
   */

  int Automaton_Homogeneous(const vector<int> & scores,
                            const int length);
 

  /** @brief build a cyclic automaton 
   *  @param cycle gives the cycle size
   *  @param final_list gives a table of positions that are final ( 0 <= final < cycle)
   *  @param final_nb   gives the size of the previous table
   *  @return 0 
   */

  int Automaton_Cycle(const int cycle, const int * const final_list, const int final_nb);

  /** @brief build a lossless automaton 
   *  @param costs gives a table of costs for each of the A letters
   *  @param max_cost is the maximal cost that can be reached so that the automaton is not in an accepting state
   *  @return 0 
   */

  int Automaton_Lossless(const vector<int> & costs, const int max_cost);
                                  

  // @}





  /** @name  Manipulate automata 
   *  @brief reduction/product/and misc probabilities
   */
  // @{
  

  /** @brief Hopcroft automaton minimization 
   *  @return the minimized automaton, does not affect the current automaton
   *  @todo{FIXME : set a compatible algorithm for "final" integer states (init at : final=0/1/2 and cross ?? )}
   */

  automaton * Hopcroft() const;

  /** @brief Isomorphism check
   *  @param  other is the automaton to be compared to
   *  @return true if the two automata are isomorph
   */

  bool       isIsomorphTo(const automaton & other) const;



  /** @brief Generic Automata Product 
   *  @param other is the second automaton used for the product
   *  @param crossFinalStates indicates if product final states are the crossproduct of both automaton final states or only one of these
   *       @li PRODUCT_UNION        : automata "union" of final states, 
   *       @li PRODUCT_INTERSECTION : automata "intersection" of final states
   *       @li PRODUCT_BUTNOT       : automata "this".final BUT NOT "other".final
   *          with
   *       @li LOOP / NO_LOOP        : indicates if the final state is a single one absorbant (force it, otherwise keep it as in the "true" product)
   *  @param otherIsProbabilist indicates that the second automaton represents a probabilistic model (false by default)
   *  @param depth indicates the maximal depth that must be reached : extra states are non final selflooping states
   *         (by default, this value is greater than 2 Billions, but the given alignment length should be enought in most cases)
   *  @return a new automaton that only gives reachable states of the product
   *  @see matrix_product
   */

  automaton * product(const automaton & other,  
                      const ProductType crossFinalStates, 
                      const bool otherIsProbabilist, 
                      const int depth) const;



  /** @name  Matrix/Matrices Products
   *  @brief matrices probabilities/cost to speed up computation (no alphabet, step product functions)
   */

#ifndef NOMATRIX
  /** @addtogroup matrix */
  // @{
#include "automaton.hh"
  // @}

  /** @brief Probabilistic/Cost matrix product of the current automaton with a probabilistic automaton
   *  @param other is the second probabilistic automaton used for the product
   *  @param crossFinalStates indicates if product final states are the crossproduct of both automaton final states or only one of these
   *       @li PRODUCT_UNION        : automata "union" of final states, 
   *       @li PRODUCT_INTERSECTION : automata "intersection" of final states
   *       @li PRODUCT_BUTNOT       : automata "this".final BUT NOT "other".final
   *          with
   *       @li LOOP / NO_LOOP        : indicates if the final state is a single one absorbant (force it, otherwise keep it as in the "true" product)
   *  @param depth indicates the maximal depth that must be reached : extra states are non final selflooping states
   *         (by default, this value is greater than 2 Billions, but the given alignment length should be enought in most cases)
   *  @return a new matrix that only gives reachable states of the product.
   *  @see product
   */
  // {@
  /// probablistic specialized product
  matrix<double> * matrix_pr_product(const automaton & other, 
                                     const ProductType crossFinalStates,
                                     const int depth) const {
    return matrix_product<double>(other,crossFinalStates,depth);
  };
  
  /// cost specialized product
  matrix<cost<int> > * matrix_cost_product(const automaton & other, 
                                           const ProductType crossFinalStates, 
                                           const int depth) const { 
    return matrix_product<cost<int> >(other,crossFinalStates,depth);
  };
  
  // }@
    
  /** @brief Probabilistic/Cost matrix slice product of the current automaton with a probabilistic automaton
   *  @brief the slice product compute the "depth" first matrices 
   *  @param other is the second probabilistic automaton used for the product
   *  @param crossFinalStates indicates if product final states are the crossproduct of both automaton final states or only one of these
   *       @li PRODUCT_UNION        : automata "union" of final states, 
   *       @li PRODUCT_INTERSECTION : automata "intersection" of final states
   *       @li PRODUCT_BUTNOT       : automata "this".final BUT NOT "other".final
   *          with
   *       @li LOOP / NO_LOOP        : indicates if the final state is a single one absorbant (force it, otherwise keep it as in the "true" product)
   *  @param depth indicates the maximal depth that must be reached 
   *  @return an array (size depth) of matrices such that each matrix gives the transition from any state to the next ...
   *  @see product
   *  @see matrix_product
   */
 
  /// probablistic specialized step product
  vector< matrix<double> * > * matrices_step_pr_product(const automaton & other, 
                                                        const ProductType crossFinalStates,
                                                        const int depth) const {
    return matrices_step_product<double>(other,crossFinalStates,depth);
  };
  
  /// cost specialized step product
  vector< matrix<cost<int> > * > * matrices_step_cost_product(const automaton & other, 
                                                              const ProductType crossFinalStates, 
                                                              const int depth) const { 
    return matrices_step_product<cost<int> >(other,crossFinalStates,depth);
  };
  
  // @}
#endif
  


  /** @brief Compute the multiple hit automaton
   *  @param m is the number of hits needed to hit an alignment
   *  @param depth indicates the maximal depth that must be reached 
   *  @return the automaton that take at least @f$m@f$ hits of the 
   *          original automaton to hit the alignment
   */
  
  automaton * mhit(const unsigned int m, 
                   const int depth) const;

  
  /** @brief Compute probability to be at a final state during the "nbSteps"th step
   *  @param nbSteps is the number of iterations done on the automaton to compute the probability 
   *  @return the probabilty to be at a final state at the "nbSteps"-th step 
   */

  double    PrFinal(const int nbSteps = gv_alignment_length) const;


  /** @brief Gives the transition probability
   *  @param a is the transition letter
   *  @param startingState is the starting state
   *  @param endingState is the ending state
   *  @return either 0 if there is no link either > 0
   */

  double    Prob(const int a, const int startingState, const int endingState) const;



  /** @brief Gives the probability to hit an alignment of the lossless set.
   *  @param nbSteps is the number of iterations done on the automaton to compute the probability
   *  @param costs is the vector of costs used on alignment alphabets
   *  @param cost_threshold is the maximal cost allowed
   *  @return the probabilty to hit such alignemnt (1.0 if lossless)
   *  @todo{FIXME : check if it does correct job for overlaps of 2 different seeds}
   *  @todo{FIXME : extends the algorithm for several hits of one/several seeds}
   *  @see Lossless 
   */

  double PrLossless( const int nbSteps, const vector<int> & costs, const int cost_threshold) const;



  /** @brief Gives the lossless property of a seed (set of seeds)
   *  @param nbSteps is the number of iterations done on the automaton to compute the probability
   *  @param costs is the vector of costs used on alignment alphabets
   *  @param cost_threshold is the maximal cost allowed
   *  @param Nocc is the minimal number of seed occurences
   *  @return a boolean that indicates if the set of seeds is lossless (at least Nocc occurences everytime),
   *  @todo{FIXME : check if it does correct job for overlaps of 2 different seeds}
   *  @todo{FIXME : extends the algorithm for several hits of one/several seeds}
   *  @see PrLossless 
   */

  bool  Lossless(const int nbSteps, const vector<int> & costs, const int cost_threshold, const int Nocc = 1) const;

    
  // @}




  /**@name  IO methods 
   * @brief IO streams to print/store/load automaton 
   */

  // @{
  /// print the graphviz/dot form of the automaton
  void dot(ostream& os) const;  
  /// print automaton information
  friend ostream& operator<<(ostream& os, const automaton& automaton); 
  /// read  automaton information
  friend istream& operator>>(istream& is, automaton& automaton);
  // @} 

  /** @name  Miscellaneous
   *  @brief miscellaneous methods
   */ 

  // @{
  ///  return the size (number of states) of the current automaton
  inline int size() const { return _states.size(); }
  ///  make the state "stateNb" loop on itselft for any letter
  inline void selfLoop(const int stateNb);
  ///  generate an alignment sequence of given length from the current probabilistic automaton
  void GenSeq(const int len) const;
  // @}


 protected:
  /** @name automaton manipulation routines 
   */
  // @{
  
  /** @brief add a new state 
   *  @param final indicates if this is a final state
   *  @return the index where the state has been added (= size() - 1)
   */
  
  inline const int addNewState(const int final = 0);


  /** @brief add a new transition between two states
   *  @param a is the transition letter read
   *  @param startingState is the initial transition state number
   *  @param endingState is the final transition state number 
   *  @param prob is the probability assigned to the current transition
   *  @return 0 
   */

  inline void addNewTransition(const int a , const int startingState , const int endingState, const double prob = (1.00/gv_align_alphabet_size));


  /** @brief change the transition probability
   *  @param "a" is the transition letter       
   *  @param startingState is the starting state
   *  @param endingState is the ending state
   *  @param prob is the new probability 
   *  @return 0 
   */

  inline int changeTransitionProb(const int a, const int startingState, const int endingState, const double prob);

  
  /** @brief check if there is a transition from the startingState labeled with "a"
   *  @param a is the transition letter
   *  @param startingState is the starting state
   *  @return true if there is at least one transition
   */

  inline bool hasTransition(const int a, const int startingState) const;
  

  // @}
  
  /// vector of states currently assigned
  vector<state> _states; 
  /// vector of init states for each cycle (if empty then always state 1 is the init state) 
  vector<int>   _init_states;
};
#endif


// @}
