/** @defgroup seed seed class
 *  @brief seed representation together with the set of cycle 
 *         positions where it should be applied 
 */

// @{

#ifndef __SEED_H__
#define __SEED_H__

//STL
#include <vector>
#include <queue>
#include <stack>
#include <list>
#include <algorithm>
//STD
#include <assert.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <string>
using namespace std;
//STR
#include "macro.h"

   
/**
 * @class seed
 * 
 * @brief represents a subset seed (can also handle a list of positions where the seed is meant to apply)
 */

class seed {
 public:

  /** @name  seed(int nbcyclespos, int maxcyclepos) 
   *  @brief build a seed of given span but full of jokers
   *  @param nbcyclepos is the number of positions in a cycle where the seed can be applied (0 = all positions),
   *  @param maxcyclepos is the size of the cycle.
   */
 
  seed(const int nbcyclepos = 0, const int maxcyclepos = 1);


  /** @name  seed(string  str_desc) 
   *  @brief build a seed according to a descriptor 
   *  @param desc is a string on ['0'-'9''A'-'Z''a'-'z']* or [(#[0-9]*)*]
   *         as "90A" or "#9#0#10" in equivalent form.
   *         if the Bsymbols option is selected, then the seed must respect 
   *         the new sysmbols definition (eg for "-spaced", "##-##-###" ...)
   */
  
  seed(string str_desc);  


 /** @name  ~seed() 
   *  @brief delete a seed and associed structures (cyclepos and seed motif)
   */  
  ~seed();


  /** @name Get functions
   *  @brief these methods are provided to get information from current seed 
   */
  
  //@{

  /** @brief gives the integer representation of the seed
   *  @return an integer table of lenght span, each element must 
   *          be in the range 0 .. |B|-1
   */
  
  int  *   table() const {return _seedMotif_int;};


  /** @brief gives the span of the seed
   *  @return the span of the seed
   */

  int      span() const  {return _span;};


  /** @brief gives the integer positions where the seed should match
   *  @return an integer table of lenght nbpos(), each element must 
   *          be in the range 0..(maxpos()-1)
   */

  int *    pos() const   {return _seedCyclePos_int;};


  /** @brief gives the number of positions where the seed should match (on a cycle)
   *  @return the number of positions
   */

  int    nbpos() const   {return _seedNbCyclePos;};


  /** @brief gives the size of the cycle
   *  @return the size of the cycle
   */

  int   maxpos()    {return _seedMaxCyclePos;};


  /** @brief  check if the seed has cycle positions (otherwise, is set at any valid position)
   *  @return true is there is cycle information associated with the current seed
   */

  bool  cycled()    {return _seedCyclePos_int != NULL;};


  /** @brief check if the seed is symetric
   *  @return true or false
   */

  bool symetric();
  
  
  /** @brief check if the seed is equal to another one
   *  @param other is the seed to be compared with this
   *  @return a boolean (only compare shapes, not positions)
   *  @todo{FIXME : do the same on position restricted seeds}
   */

  bool equal(seed * other);


  /** @brief gives the current seed shape (a string
   *         in [0-9A-Za-z]* or (#[0-9]*)* with a possible position restricted part ":1,3,5/7"
   *  @return a string object that represents this seed.
   */

  string str();

  
  //@}


  //@{
  
  /** @brief compute the weight of a given seed
   *  @return the weight as a double value
   */

  double weight();

  
  /** @brief compute seed selectivity according to the weight (bernoulli model estimation)
   *  @return the selectivity as a double value
   */

  double selectivityFromWeight();


  /** @brief check if the current seed is acceptable : 
   * this implies a correct weight  (according to gloval variables gv_minweight and gv_maxweight), 
   * and a symetric shape when requested.
   *  @return true when the current seed is acceptable
   */
  bool acceptable();


  //@}



  //@{

  /** @brief check if the signature is realisable according to the span : stop the program otherwise
   */
  
  void checksignature();


  /** @brief set the very first seed signature
   */

  void setsignature();


  //@}



  //@{

  /** @brief set all the elements of position >= first_pos that are <= b in the seed in the decreasing order
   *  @brief this function is used to enumerate signatures and inner used.
   *  @param first_pos gives the first position to be sorted
   *  @param b gives the maximal value that can be sorted (all others remain unchanged)
   *  @see next()
   */

  void reorder(int first_pos,int b);


  /** @brief generates the next seed 
   *  @brief this function does an enumerative generation of the acceptable seeds (and the cycle pos when required), 
   *         with or without signature constraits
   *  @return 1 if next seed has been generated, 0 if all the seeds in the set have already been generated
   *  @see acceptable()
   *  @see setsignature(),checksignature()
   */

  int next();


  /** @brief this function generates a random seed (and random positions) according to gv_minspan and gv_maxspan 
   *
   *  @return 1 all the time
   */

  int random();


  /** @brief does a swap for the hill climbing function (swap cycle positions then pair of seeds elements, one at a time)
   *  @return 1 if the next swap has been generated, 0 if all possible swaps have already been enumerated
   */
  
  int nextswap();


  /** @brief reset the swap for the hill climbing function, all swap will then be processed from the begining
   */

  void resetswap();


  /** @brief fix definitely the swap (reverse it)
   */

  void setswap();


  //@}



  //@{

  /** @brief set cycle properties of the seed 
   *  @param cycle_pos is a vector of positions where the seed is allowed to match
   *  @param cycle_size is the maximal position where a position can be set
   */

  void setCyclePos(vector<int> cycle_pos, int cycle_size);

  
  /** @brief output a seed object
   *  @param os is the outputstream
   *  @param s is the seed to output
   *  @return the outputsteam
   *  @see seed
   *  @todo{FIXME : implement also the input operator for a seed (set of seeds ?) with cycle reading and strong parsing}
   */

  friend ostream& operator<<(ostream& os, seed& s);


 protected:
  /// Current seed motif represented as an int array {0,1,2, ...} of lenght "_span"
  int  * _seedMotif_int;
  /// Fixed seed span 
  int _span;
  /// Swapped char cursor (from) 
  int _swap_from;
  /// Swapped char cursor (to)
  int _swap_to;


  /// Current seed positions to be set (NULL if applied on all positions)
  int * _seedCyclePos_int;
  /// Size of the "_seedCyclePos_int" table
  int _seedNbCyclePos;
  /// Max value in the  "_seedCyclePos_int" table
  int _seedMaxCyclePos;

  /// Swapped pos index (0 if no swap)
  int _swap_pos_index;
  /// Swapped old value (0 if no swap)
  int _swap_pos_old_value;



  /** 
   * @brief convert a char into a number
   * @brief '0' -> 0, '1' -> 1 ... '9' -> 9, 'A' -> 10 ... 'Z' -> 35, 'a' -> 36 ...
   * @param s is the string where conversion is currently processed
   * @param pos is the first char inside s where the conversion has to be done
   * @return the converted value (integer that represents a seed letter) and increments i to the next symbol
   */

  int convert(string & s, int * pos);


  /** 
   * @brief convert an integer into a string
   * @brief 0 -> '0', 1 -> '1' ... 9 -> '9', 10 -> 'A' ... 35 -> 'Z', 36 -> 'a' ...
   * @param i is the integer currently processed
   * @return the converted value (string that represents a seed letter)
   */

  string rconvert(int i);

};

#endif

// @}
